from typing import List, Any
from typing import Union

from ....yaml_structures.yaml_dictionary import YamlDictionary
from ....yaml_structures.yaml_list import YamlList

class ValueByPathBooleanSearcher:
    @staticmethod
    def search(filters: List[str], value: Union[str, int], search_in_objects: list) -> bool:
        return ValueByPathBooleanSearcher._search_recursion(filters, value, search_in_objects, True)
    
    @staticmethod
    def _search_recursion(filters: List[str], value: Union[str, int], search_in_objects: list, returned_value: bool) -> bool:
        if len(filters) == 0:
            if len(search_in_objects) > 0 :
                first_search_object_str_int: str = search_in_objects.pop(0)
                if ((isinstance(first_search_object_str_int, str) or isinstance(first_search_object_str_int, int)) and 
                        (first_search_object_str_int == value)):
                    return True
                else:
                    return False
                
            return returned_value
        else:
            first_filter: str = filters.pop(0)
            first_search_object: Union[YamlDictionary, YamlList ] = search_in_objects.pop(0)
            if first_filter != "[]":
                if isinstance(first_search_object, YamlList):
                    raise Exception
                
                if isinstance(first_search_object, list):
                    sub_returned_value: bool = False
                    for sub_item in first_search_object:
                        sub_returned_value = sub_returned_value or ValueByPathBooleanSearcher._search_recursion([first_filter] + filters.copy(), value, [sub_item], True)
                    
                    return ValueByPathBooleanSearcher._search_recursion([], value, [], returned_value and sub_returned_value)
                
                elif isinstance(first_search_object, YamlDictionary) and first_search_object.key == first_filter:
                    
                    return ValueByPathBooleanSearcher._search_recursion(filters, value, [first_search_object.value], returned_value and True)
                else:
                    
                    return ValueByPathBooleanSearcher._search_recursion(filters, value, [first_search_object], returned_value and False)
                                  
            else:
                if first_filter == "[]":
                    if not isinstance(first_search_object, YamlList):
                        raise Exception
               
                return ValueByPathBooleanSearcher._search_recursion(filters, value, first_search_object.values, returned_value and True) # type: ignore


from yaml import dump as upload_data # type: ignore
from yaml import Dumper
from yaml import YAMLError
from typing import Any, List, Union, cast
from ..translators.exceptions.rule_not_found_exception import RuleNotFoundException
from ..translators.exceptions.writing_yaml_exception import WritingYamlException
from ..yaml_structures.yaml_dictionary import YamlDictionary
from ..yaml_structures.yaml_list import YamlList
from ...utils.file.file_utils import FileUtil


class ToYamlTranslator:
    """Translator from the format of Yaml wrapper, to the format of yaml library.
    """
    def __init__(self , file_path: str) -> None:
        self._file_path: str = file_path
        
    def translate(self, content_to_translate: list) -> Union[dict, list, None]:
        """Translates the content of the YamlWrapper to the format of yaml library.

        Returns:
            Union[dict, list, None]: Content translated in the format accepted by yaml library.
            
        Note:
            Temporary the support for lists is only for list with integers and strings, and lists of lists of YamlDictionary.
        """
        if len(content_to_translate) == 0:
            FileUtil.delete_file(self._file_path)
            FileUtil.create_file_from_path(self._file_path)
            return None
        else:    
            content_to_write: Union[dict, list] = self._apply_rule_rec(content_to_translate, {})
            self._write_to_file(content_to_write)
            return content_to_write
    
    def _write_to_file(self, content_to_write: Union[dict, list]) -> None:
        try:
            with open(self._file_path, "w") as file:
                upload_data(content_to_write, file, Dumper)
        except YAMLError:
            raise WritingYamlException("Error writing the content to the yaml file.")
        
    def _apply_rule_rec(self, objects: Union[List[YamlDictionary], List[YamlList], Any], returned_value: Union[dict, list]) -> Union[dict, list]:
        if len(objects) == 0: 
            return returned_value
        else:
            first_object: Union[YamlDictionary, YamlList] = objects[0]
            value_to_return: dict = cast(dict,returned_value)
            
            if isinstance(first_object, YamlDictionary):
                if isinstance(first_object.value, str) or isinstance(first_object.value, int):
                    
                    value_to_return[first_object.key] = first_object.value 
                    
                elif isinstance(first_object.value, list) and isinstance(first_object.value[0], YamlDictionary):
                   
                    value_to_return[first_object.key] = self._apply_rule_rec(first_object.value, {})
                    
                else:
                    value_to_return[first_object.key] = self._apply_rule_rec([first_object.value], [])

                return  self._apply_rule_rec(objects[1:], value_to_return)
            else:
                if isinstance(first_object, YamlList):
                    if isinstance(first_object.values[0], int) or isinstance(first_object.values[0], str):
                        return first_object.values
                    else: 
                        partial_return: list = []
                        for item in  first_object.values:
                            partial_return.append(self._apply_rule_rec(item, {}))
                        return partial_return
        


from typing import Any, List, Union, Tuple
from ...translators.enums.object_type import ObjectType
from ...translators.enums.rule_type import ListRuleType, RuleType
from ...translators.exceptions.rule_not_found_exception import RuleNotFoundException
from ...yaml_structures.yaml_dictionary import YamlDictionary
from ...yaml_structures.yaml_list import YamlList


class RuleFromValueUtil:
  
    @staticmethod
    def define_rule_from_value(value: Any) -> RuleType:
        if isinstance(value, int):
            return RuleType.INT_RULE
        elif isinstance(value, str):
            return RuleType.STR_RULE
        elif isinstance(value, dict):
            return RuleType.LIST_DICT_RULE
        elif isinstance(value, list):
            return RuleType.LIST_RULE
        else:
            raise RuleNotFoundException("Not able to find a valid rule.")
        
    @staticmethod
    def define_list_rule_type_from_value(value: List[Union[dict, Any]]) -> ListRuleType:
            first_element: Union[dict, Any ] = value[0]
            if isinstance(first_element, dict):
                return ListRuleType.DICT_TYPE
            
            return ListRuleType.OTHER_TYPE
        

class RuleFromObjectUtil:
    
    @staticmethod
    def define_rule_from_object(object: Union[YamlDictionary, YamlList]) -> ObjectType:
        if isinstance(object, YamlDictionary):
            return ObjectType.DICT
        elif isinstance(object, YamlList):
            return ObjectType.LIST
        else:
            raise RuleNotFoundException("Not able to find a valid rule.")




from typing import List, Any
from typing import Union
from typing import cast
from ....yaml_structures.exceptions.list_not_respect_integrity import ListNotRespectIntegriry
from ....yaml_structures.yaml_dictionary import YamlDictionary
from ....yaml_structures.yaml_list import YamlList

class ObjectByPathReference:
    @staticmethod
    def search(filters: List[str], search_in_objects: List[Any]) -> Union[YamlDictionary, YamlList, None]:
        return ObjectByPathReference._search_recursion(filters, search_in_objects, cast(Union[YamlDictionary, YamlList], {}))
    
    @staticmethod
    def _search_recursion(filters: List[str], search_in_objects: list, returned_value: Union[YamlDictionary, YamlList, None]) -> Union[YamlDictionary, YamlList, None]:
        if len(filters) == 0:
            return returned_value
        else:
            for item in search_in_objects:
                first_filter: str = filters[0]
                
                if (isinstance(item, YamlDictionary) and 
                    cast(YamlDictionary, item).key == first_filter):
                    
                    if(len(filters[1:]) > 0) and  ((not isinstance(item.value, int)) or (not isinstance(item.value, int))):
                        return ObjectByPathReference._search_recursion(filters[1:], cast(list, item.value), None)
                    else:
                        search_in_objects.pop(search_in_objects.index(item))
                        return ObjectByPathReference._search_recursion([], [], item)
                    
            return  None